#   Copyright (c) 2021 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

import subprocess
from datetime import datetime
import logging
import xmltodict
from collections import OrderedDict

from webservice_fdsn.app.core import constants
from webservice_fdsn.app.utils import pf_helper
from webservice_fdsn.app.utils.exceptions import UnsupportedDBSchema
from webservice_fdsn.app.utils import dbcluster_resolver
from webservice_fdsn.app.core.event_exporter.export_events_helpers import exceptions

log = logging.getLogger(__name__)


def get_db_schema(dbpath):
    """Determine schema of db at path dbpath

    :param str dbpath: path to database file
    :rtype: str
    """
    # Python datascope interface does not have dbquery yet, have to resort
    # to subprocess
    cmd = ['dbquery', dbpath, 'dbSCHEMA_NAME']
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    schema = p.communicate()[0]
    log.debug(f'Got db schema as {schema}')
    if not schema:
        raise UnsupportedDBSchema(f'Unable to fetch DB schema')
    if b'0.5' in schema:
        return constants.dbcluster_constants.DBCLUSTER05
    elif b'0.7' in schema:
        return constants.dbcluster_constants.DBCLUSTER07
    elif b'css' in schema:
        # Strip any newlines/spaces
        return schema.decode('utf-8').strip()
    else:
        raise UnsupportedDBSchema(f'DB schema {schema.decode("utf-8")} is not supported')


def get_db_files(caller, scratch_dir, starttime, endtime):
    """Get database file/s on disk

    If regular css3.0, returns a single file. If dbcluster schema, then multiple
    files will be returned depending on how many files the start and end times span.

    :param str caller: which module is calling this helper method
    :param str scratch_dir: path to scratch directory to be used by this service
    :param datetime.datetime starttime: start time of window to get data
    :param datetime.datetime endtime: end time of window to get data
    :rtype list
    :return: list of databases
    """
    db_path = pf_helper.get_project_database_path()

    # TODO: db schema does not change thru the lifetime of the app, so need a way
    # to detect and persist schema information at app start-up time instead
    # of doing this at every request
    # Check db schema.. could be dbcluster or regular cssx.x
    schema = get_db_schema(db_path)

    if schema not in [constants.dbcluster_constants.DBCLUSTER07,
                      constants.dbcluster_constants.DBCLUSTER05]:
        log.debug('css db detected')
        return [db_path]

    log.debug(f'dbcluster detected with schema {schema}')

    # create a DbClusterResolver object, and get_dfiles_by_date files
    # combine them and pass it as db_path to get_event_data method below
    # Detect whether requested time window is contained in the dbcluster first though,
    # and exit early if not.
    dbcr = dbcluster_resolver.DBClusterResolver(db_path)
    dbs = dbcr.get_dfiles_by_date(caller, starttime,
                                  endtime, sanitize=True)
    return dbs


def handle_limit_and_offset(evids, limit, offset):
    """Apply limit and offset constraints to evid list

    :param list evids: list of event ids
    :param int limit: number of event ids to return
    :param int offset: starting index of event ids to return from
    :rtype list
    :raises exceptions.NoEventDataFound: If no events left after applying limit/offset
    :return: list of remaining event IDs
    """
    log.debug(f'Applying limit[{limit}] and offset[{offset}] to received evid list of length {len(evids)}')

    if not isinstance(evids, list):
        raise Exception(f'Expected evids to be of type list, got type {type(evids)} instead')

    if limit == 0:
        raise exceptions.NoEventDataFound('No events left after applying limit and offset')
    if offset:
        evids = evids[offset:]
    if limit and evids:
        evids = evids[:limit]

    log.debug(f'Left with {len(evids)} events after applying limit and offset')
    if not evids:
        raise exceptions.NoEventDataFound('No events left after applying limit and offset')
    return evids


def combine_quakemls(quakemls, limit=None, offset=None):
    """Combine multiple quakeml defintions into a single doc.

    :param list quakemls: list of quakeml docs
    :param int offset: offset to return events from
    :param int limit: number of events to return
    :rtype: str
    :return: combined doc
    """
    # Iterate through each entry, convert to python dict object using xmltodict
    # Then aggregate the events, and convert resultant object back to quakeml
    # using xmltodict.unparse
    if not quakemls:
        return ''

    log.debug(f'Combining {len(quakemls)} documents. Limit: {limit}, offset: {offset}')

    # Get the parent which will collect all events
    parent_qml_meta = xmltodict.parse(quakemls[0])
    log.debug(f'Before merge: Parent has {len(parent_qml_meta["q:quakeml"]["eventParameters"]["event"])} events')

    # HACK - to be investigated
    # Check datatype of 'event' member. event2qml sends back event data as an OrderedDict is single event,
    # else it is sent back as a list. Convert to list before proceeding.
    # Not sure of what the standard is for quakeml, but ideally it would be great for us if event2qml always sends
    # evids as a list irrespective of number of events, so this type check can be avoided.
    if isinstance(parent_qml_meta['q:quakeml']['eventParameters']['event'], OrderedDict):
        log.debug('evids came in as type OrderedDict, converting to list')
        parent_qml_meta['q:quakeml']['eventParameters']['event'] = [parent_qml_meta['q:quakeml']['eventParameters']['event']]

    # Aggregate events from remaining quakemls
    for qml in quakemls[1:]:
        xd = xmltodict.parse(qml)
        evnts = xd['q:quakeml']['eventParameters']['event']
        # HACK: To be investigated, evnts should always be a list as retured by event2qml module
        if isinstance(evnts, OrderedDict):
            log.debug('evnts is of type OrderedDict, converting to list')
            evnts = [evnts]
        log.debug(f'Appending {len(evnts)} events to parent')
        parent_qml_meta['q:quakeml']['eventParameters']['event'].extend(evnts)

    log.debug(f'After merge: Parent has a total of {len(parent_qml_meta["q:quakeml"]["eventParameters"]["event"])} aggregated events')

    # Apply limit and offset on the final data set
    parent_qml_meta["q:quakeml"]["eventParameters"]["event"] = \
        handle_limit_and_offset(parent_qml_meta["q:quakeml"]["eventParameters"]["event"],
                                limit, offset)
    return xmltodict.unparse(parent_qml_meta)
