#   Copyright (c) 2020 Boulder Real Time Technologies, Inc.
#
#   This software module is wholly owned by Boulder Real Time
#   Technologies, Inc. Any use of this software module without
#   express written permission from Boulder Real Time Technologies,
#   Inc. is prohibited.
#
#   Author: Ambli LLC (ambli.ll.company@gmail.com)

import os
import logging
import tempfile
import connexion
from flask import make_response
from datetime import datetime, timezone, timedelta
from dateutil import parser
from flask import send_file
from http import HTTPStatus

from webservice_fdsn.app.v1_1_b.models.data_select import DataSelect  # noqa: E501
from webservice_fdsn.app.v1_1_b import util
from webservice_fdsn.app.core.waveform_helpers import waveform_queries, exceptions as wf_helper_exceptions
from webservice_fdsn.app.utils import pf_helper
from webservice_fdsn.app.core import constants
from webservice_fdsn.app.utils import exceptions as apiutil_exceptions
from webservice_fdsn.app.utils import utilities


log = logging.getLogger(__name__)


def get_data_select_wadl():  # noqa: E501
    """get_data_select_wadl

     # noqa: E501


    :rtype: None
    """
    wadl_path = os.path.join(os.path.dirname(__file__), os.pardir, "wadl", "dataselect.wadl")
    with open(wadl_path, "r") as f:
        resp = make_response("".join(f.readlines()), HTTPStatus.OK)
        return resp


def dataselect_query_handler(starttime: datetime, station: str, channel: str, location: str, endtime: datetime, nodata: int):
    """Service dataselect query per user requested parameters
    :param datetime.datetime starttime:
    :param str station:
    :param str channel:
    :param str location:
    :param datetime.datetime endtime:
    :param int nodata
    :return
    """
    # Get path to db
    try:
        db_path = pf_helper.get_project_database_path()
    except KeyError:
        _msg = "Service configuration error: Path to Antelope DB must be configured in parameter file"
        log.error(_msg)
        resp = make_response(_msg, HTTPStatus.INTERNAL_SERVER_ERROR)
        return resp

    scratch_dir_path = pf_helper.get_project_scratch_dir()

    # handle_dataselect_queries relies on trexcerpt to extract data from db. Unfortunately trexcerpt does not
    # support filtering data using network information, only station, channel, starttime and endtime.
    # Per Kent, most servers hosting the Antelope FDSN webservice initially will support data from only 1 network,
    # so network value part of webservice pf file which configures the webservice and queries can be constructed
    # assuming this value for the network parameter.
    # Possibility of extending trexcerpt to support network parameters can be considered at a later date.
    # Fetch preset network value from pf file
    service_network = pf_helper.get_default_service_network()

    # Get database file/s on disk
    try:
        db_paths = utilities.get_db_files(constants.service_interfaces_constants.DATASELECT,
                                          scratch_dir_path, starttime, endtime)
    except (apiutil_exceptions.RequiredPfOptionMissing,
            apiutil_exceptions.InvalidPfValue) as ex:
        log.error(f"Service configuration error: {ex}")
        resp = make_response(f'{ex}', HTTPStatus.INTERNAL_SERVER_ERROR)
        return resp

    if not db_paths:
        resp = make_response("", nodata)

    wf_paths = []
    for db_path in db_paths:
        try:
            wf_path = waveform_queries.handle_dataselect_query(db_path, service_network,
                                                               station, channel,
                                                               "--", starttime,
                                                               endtime, scratch_dir_path)
        except wf_helper_exceptions.RequestedDataNotFoundError as ex:
            log.warning(f'User requested data not found, message: {ex}')
            wf_path = None
        except wf_helper_exceptions.InvalidInput as ex:
            log.error(f'User passed in bad input, error: {ex}')
            return make_response(format(ex), HTTPStatus.BAD_REQUEST)
        except wf_helper_exceptions.ResponseTooLarge as ex:
            log.warning(f'User passed in bad input, error: {ex}')
            return make_response(format(ex), HTTPStatus.FORBIDDEN)
        except wf_helper_exceptions.InvalidAppConfig as ex:
            _msg = f'Invalid application config detected, contact your service administrator with this error message: {ex}'
            log.error(_msg)
            return make_response(_msg, HTTPStatus.INTERNAL_SERVER_ERROR)
        else:
            if wf_path:
                wf_paths.append(wf_path)

    # each member in wf_paths is the mseed data. Need to combine all files into a single one
    # and return that to the user. So this logic is correct, need to implement and test with
    # dbcluster waveform data
    if wf_paths:
        log.debug(f'Merging {wf_paths}')
        with tempfile.NamedTemporaryFile(dir=scratch_dir_path, delete=True,
                                         suffix='.mseed', prefix='fdsnws-dataselect-') as f:
            waveform_queries.merge_files(wf_paths, f.name)
            return send_file(f.name, as_attachment=True)
    else:
        resp = make_response("", nodata)
        return resp


def get_dataselect(starttime, station, channel, location, endtime=None, nodata=None):  # noqa: E501
    """:return:

     # noqa: E501

    :param starttime: Limit results to time series samples on or after the specified start time (UTC)
    :type starttime: str
    :param station: Select one or more SEED station codes. Multiple codes are comma-separated.
    :type station: str
    :param channel: Select one or more SEED channel codes. Multiple codes are comma-separated.
    :type channel: str
    :param location: Select one or more SEED location identifiers. Multiple identifiers are comma- separated. As a 
    special case "--" (two dashes) will be translated to a string of two space characters to match blank location IDs.
    :type location: str
    :param endtime: Limit results to time series samples on or before the specified end time (UTC)
    :type endtime: str
    :param nodata: Select status code for "no data", either 204 (default) or 404
    :type nodata: str

    :rtype: None
    """
    log.info(f'Querying waveform data using params: starttime: {starttime}, station: {station},'
             f' channel: {channel}, location: {location}, endtime: {endtime}')
    try:
        starttime = parser.parse(starttime)
    except ValueError as ex:
        resp = make_response('Invalid starttime provided', HTTPStatus.BAD_REQUEST)
        return resp
    if endtime:
        try:
            endtime = parser.parse(endtime)
        except ValueError:
            resp = make_response('Invalid endtime provided', HTTPStatus.BAD_REQUEST)
            return resp
    else:
        # If endtime not specified, return 1 hour of waveform data
        log.debug("No end time specified, setting it to 1 hour after start time")
        endtime = starttime + timedelta(hours=1) 

    return dataselect_query_handler(starttime, station, channel, location, endtime, nodata)


def get_dataselect_version():  # noqa: E501
    """get_dataselect_version

     # noqa: E501


    :rtype: None
    """
    resp = make_response("1.1.0", HTTPStatus.OK)
    return resp


def post_dataselect(body):  # noqa: E501
    """:return:

     # noqa: E501

    :param body: 
    :type body: dict | bytes

    :rtype: None
    """
    if connexion.request.is_json:
        body = DataSelect.from_dict(connexion.request.get_json())  # noqa: E501

    starttime = body.starttime.replace(tzinfo=timezone.utc)
    if body.endtime:
        endtime = body.endtime.replace(tzinfo=timezone.utc)
    else:
        endtime = None
    return dataselect_query_handler(starttime,
        body.station,
        body.channel,
        body.location,
        endtime,
        body.nodata)
